// SJJ Embedded Micro Solutions, LLC.
// Copyright  2009 SJJ Embedded Micro Solutions, LLC. All Rights Reserved
//  
// THIS SAMPLE CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.

/*****************************************************************************
*
* FILE NAME:    Sema4Mutex.c
*
* DESCRIPTION:    This is the main program module.
*
\*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <rt.h>


#include "Sema4Mutex.h"

// global variables
  RTHANDLE hRootProcess;
  DWORD    dwKtickInUsecs;

//++
  char cSharedBuff[MAX_BUFF];

  // Mutual exclusion semaphore handle
  RTHANDLE  hMutex;
//++

// mailbox and semaphore thread declarations

  void     ThreadWriter(void);

  void     ThreadReader(void);

// module variables
struct _INIT {
  enum States
       { BEFORE_INIT,
         INIT_BUSY,
         INIT_DONE,
         CLEANUP_BUSY } state;
  RTHANDLE hMain;
  BOOL     bCataloged;
	// mailbox and semaphore thread declarations

  RTHANDLE hThreadWriter;

  RTHANDLE hThreadReader;

}  strInit;

static void     Cleanup(void);

/*****************************************************************************
*
* FUNCTION:     main
*
* DESCRIPTION:
*  This is the main program module.
*  It creates shared memory objects, and all threads.
*  The main thread then waits for a deletion message,
*  and then all objects are deleted and the process terminates itself.
*
\*****************************************************************************/

void        main(void)
{
  SYSINFO   sysinfo;
  EVENTINFO eiEventInfo;

//++
  BYTE bThreadPriority;
//++

#ifdef _DEBUG
  printf("Sema4Mutex started\n");
#endif

  // obtain handle of root process (cannot fail)
  hRootProcess = GetRtThreadHandles(ROOT_PROCESS);

  // initialize the structure for cleaning up
  memset(&strInit, 0, sizeof(strInit));
  strInit.state = BEFORE_INIT;
  
  // get low level tick length in usecs
  if (!CopyRtSystemInfo(&sysinfo))
    Fail("Cannot copy system info");
  dwKtickInUsecs = 10000 / sysinfo.KernelTickRatio;
  if (dwKtickInUsecs == 0)
    Fail("Invalid low level tick length");

  // adjust process max priority (ignore error)
  SetRtProcessMaxPriority(NULL_RTHANDLE, 150);

  // obtain main thread's handle
  strInit.hMain = GetRtThreadHandles(THIS_THREAD);
  strInit.state = INIT_BUSY;

//++
#ifdef _DEBUG
  printf("Main Process Thread Handle: 0x%x\n", strInit.hMain);
#endif
//++

  // catalog the token of this process in the root process
  if (!Catalog(hRootProcess,
      GetRtThreadHandles(THIS_PROCESS),
      "Sema4Mutex"))
    Fail("Cannot catalog process name");
  strInit.bCataloged = TRUE;

//++
  // Create & catalog the mutual exclusion semaphore
  hMutex = CreateRtSemaphore(0, 1, FIFO_QUEUING);
  if (hMutex == BAD_RTHANDLE)
  {
    Fail("Cannot create semaphore hMutex");
  }
  if (!Catalog(NULL_RTHANDLE, hMutex, "Mutex"))
  {
    Fail("Cannot catalog semaphore hMutex");
  }

  // Zero shared buffer
  cSharedBuff[0] = 0;

  // Determine thread priority for Reader and Writer threads
  // Get current thread's priority
  bThreadPriority = GetRtThreadPriority(NULL_RTHANDLE);

  // Make Reader and Writer thread priority higher than the current thread
  bThreadPriority--;
//++

  // create mailbox and semaphore threads
//+-
  strInit.hThreadWriter
          = CreateRtThread(bThreadPriority,
            ThreadWriter,
            4096,
            0);
//+-
  if (strInit.hThreadWriter == BAD_RTHANDLE)
  {
    strInit.hThreadWriter = NULL_RTHANDLE;
    Fail("Cannot create thread for Writer");
  }

//++
#ifdef _DEBUG
  printf("Writer Thread Handle: 0x%x\n", strInit.hThreadWriter);
#endif
//++

//+-
  strInit.hThreadReader
          = CreateRtThread(bThreadPriority,
            ThreadReader,
            4096,
            0);
//+-

  if (strInit.hThreadReader == BAD_RTHANDLE)
  {
    strInit.hThreadReader = NULL_RTHANDLE;
    Fail("Cannot create thread for Reader");
  }

//++
#ifdef _DEBUG
  printf("Reader Thread Handle: 0x%x\n", strInit.hThreadReader);
#endif
//++

  // indicate that initialization has finished
  strInit.state = INIT_DONE;
#ifdef _DEBUG
  printf("Sema4Mutex finished initialization\n");
#endif

//++
  printf("Releasing mutual exclusion semaphore\n");
  if ( !ReleaseRtSemaphore(hMutex, 1) )
  {
    Fail("Main process thread failed to release semaphore!");
  }
//++

  // wait for notification
  while (RtNotifyEvent(RT_SYSTEM_NOTIFICATIONS | RT_EXIT_NOTIFICATIONS,
      WAIT_FOREVER,
      &eiEventInfo))
  {

    switch(eiEventInfo.dwNotifyType)
    {
    case TERMINATE:
      // TODO: this process should terminate
      // cleanup the environment
      Cleanup();  // does not return

    case RT_CLIENT_DOWN:
      // TODO: on a NT host react to a RT client that has gone down
      break;

    case RT_CLIENT_UP:
      // TODO: on a NT host react to a RT client that has come back
      break;

    case NT_HOST_DOWN:
      // TODO: on a RT client react to a NT host that has gone down
      break;

    case NT_HOST_UP:
      // TODO: on a RT client react to a NT host that has come back
      break;

    case NT_BLUESCREEN:
      // TODO: react to a NT blue screen
      break;
    }
  }
  Fail("Notify failed");
}


/*****************************************************************************
*
* FUNCTION:     Fail
*
* PARAMETERS:   same parameters as expected by printf
*
* RETURNS:      does not return
*
* DESCRIPTION:
*  If in debug mode, prints the message.
*  Then the current process is killed ungraciously.
\*****************************************************************************/

void        Fail(
  LPSTR       lpszMessage, ...)
{
  EXCEPTION   eh;
  RTHANDLE    hDelMbx;
  DWORD     dwTerminate;

#ifdef _DEBUG
  va_list     ap;

  va_start(ap, lpszMessage);
  vprintf(lpszMessage, ap);
  va_end(ap);
  printf("\nError nr=%x\n", GetLastRtError());
#endif

  // make sure that exceptions are returned for inline handling
  GetRtExceptionHandlerInfo(THREAD_HANDLER, &eh);
  eh.ExceptionMode = 0;
  SetRtExceptionHandler(&eh);

  // if we had not started initializing yet, just get out
  if (strInit.state == BEFORE_INIT)
    exit(0);

  if (strInit.hMain == GetRtThreadHandles(THIS_THREAD))
  {
    // this is the main thread:
    // if we are busy initializing, then do Cleanup
    if (strInit.state == INIT_BUSY)
      Cleanup();  // does not return

    // this is the main thread, but we are not initializing:
    // just return
    return;
  }

  // this is not the main thread:
  // ask main thread to do cleanup
  // (allow some time to setup the deletion mailbox, ignore errors)
  hDelMbx = LookupRtHandle(NULL_RTHANDLE, "R?EXIT_MBOX", 5000);
  dwTerminate = TERMINATE;
  SendRtData(hDelMbx, &dwTerminate, 4);
  SuspendRtThread(NULL_RTHANDLE);

  // if we get here, suspend has failed
  while (1)
    RtSleep(655349);  // sleep for the maximum time
}


/*****************************************************************************
*
* FUNCTION:   Cleanup (local function)
*
* DESCRIPTION:
*  Remove all objects and other process side effects, 
*  as far as they have been created.
*
\*****************************************************************************/

void      Cleanup(void)
{
  // indicate that we are cleaning up
  strInit.state = CLEANUP_BUSY;

  // kill mailbox and semaphore threads

  if (strInit.hThreadWriter != NULL_RTHANDLE)
    if (!DeleteRtThread(strInit.hThreadWriter))
      Fail("Cannot delete thread for Writer");

  if (strInit.hThreadReader != NULL_RTHANDLE)
    if (!DeleteRtThread(strInit.hThreadReader))
      Fail("Cannot delete thread for Reader");

//++
  // Uncatalog semaphore handle and delete semaphore
  if (hMutex != BAD_RTHANDLE)
  {
    if (!UncatalogRtHandle(NULL_RTHANDLE, "Mutex"))
    {
      Fail ("Uncatalog of Mutex failed!");
    }
    if (!DeleteRtSemaphore(hMutex))
    {
      Fail ("Mutex semaphore delete failed");
    }
  }
  else
  {
    Fail ("Cannot cleanup mutex semaphore");
  }
//++

  // remove our name from the root process
  if (strInit.bCataloged)
    if (!UncatalogRtHandle(hRootProcess, "Sema4Mutex"))
      Fail("Cannot remove my own name");

#ifdef _DEBUG
  printf("Sema4Mutex finished cleaning up\n");
#endif

  // lie down
  exit(0);
}
